@@ -19,11 +19,11 @@ from typing import Tuple, Union
 
 import numpy as np
 
-__version__ = '1.0.0'
+__version__ = '1.1.0'
 
 
 #
-# S2 base constants needed for S2 cell mapping
+# S2 base constants needed for cell mapping
 #
 
 # The maximum level supported within an S2 cell ID. Each level is represented by two bits in the
@@ -69,7 +69,8 @@ _S2_POS_TO_IJ = np.array([
     [3, 1, 0, 2],  # 3: Swap and invert bits set
 ], dtype=np.uint64)
 
-# Lookup for the orientation update mask of one of the four sub-cells within a higher level cell
+# Lookup for the orientation update mask of one of the four sub-cells within a higher level cell.
+# This mask is XOR'ed with the current orientation to get the sub-cell orientation.
 _S2_POS_TO_ORIENTATION_MASK = np.array([
     _S2_SWAP_MASK, 0, 0, _S2_SWAP_MASK | _S2_INVERT_MASK
 ], dtype=np.uint64)
@@ -195,7 +196,82 @@ def _s2_init_lookups() -> None:
 
 
 #
-# Public functions
+# Cell ID <-> Token translation functions
+#
+
+def cell_id_to_token(cell_id: Union[int, np.uint64]) -> str:
+    """
+    Convert S2 cell ID to a S2 token.
+
+    Converts the S2 cell ID to hex and strips any trailing zeros. The 0 cell ID token is represented
+    as 'X' to prevent it being an empty string.
+
+    See s2geometry/blob/c59d0ca01ae3976db7f8abdc83fcc871a3a95186/src/s2/s2cell_id.cc#L204-L220
+
+    Args:
+        cell_id: The S2 cell ID integer.
+
+    Returns:
+        The S2 token string for the S2 cell ID.
+
+    Raises:
+        TypeError: If the cell_id is not int or np.uint64.
+
+    """
+    # Check type
+    if not isinstance(cell_id, (int, np.uint64)):
+        raise TypeError('Cannot convert S2 cell ID from type: ' + str(type(cell_id)))
+    cell_id = np.uint64(cell_id)
+
+    # The zero token is encoded as 'X' rather than as a zero-length string. This implementation has
+    # no method of generating the 0 cell ID, so this line is mostly here for consistency with the
+    # reference implementation.
+    if cell_id == 0:  # pragma: no cover
+        return 'X'
+
+    # Convert cell ID to hex and strip any trailing zeros
+    return '{:016x}'.format(cell_id).rstrip('0')
+
+
+def token_to_cell_id(token: str) -> np.uint64:
+    """
+    Convert S2 token to S2 cell ID.
+
+    Restores the stripped 0 characters from the token and converts the hex string to integer.
+
+    See s2geometry/blob/c59d0ca01ae3976db7f8abdc83fcc871a3a95186/src/s2/s2cell_id.cc#L222-L239
+
+    Args:
+        token: The S2 token string. Can be upper or lower case hex string.
+
+    Returns:
+       The S2 cell ID for the S2 token.
+
+    Raises:
+        TypeError: If the token is not str.
+        ValueError: If the token length is over 16.
+
+    """
+    # Check input
+    if not isinstance(token, str):
+        raise TypeError('Cannot convert S2 token from type: ' + str(type(token)))
+
+    if token in ('x', 'X'):
+        # The zero cell ID is represented as the character 'X' rather than as an empty string
+        return np.uint64(0)
+
+    if len(token) > 16:
+        raise ValueError('Cannot convert S2 token with length > 16 characters')
+
+    # Add stripped zeros
+    token = token + ('0' * (16 - len(token)))
+
+    # Convert to cell ID by converting hex to int
+    return np.uint64(int(token, 16))
+
+
+#
+# Encode functions
 #
 
 def lat_lon_to_cell_id(  # pylint: disable=too-many-locals
@@ -378,18 +454,13 @@ def lat_lon_to_token(lat: float, lon: float, level: int = 30) -> str:
         ValueError: When level is not an integer, is < 0 or is > 30.
 
     """
-    # Get the cell ID for the lat/lon
-    cell_id = lat_lon_to_cell_id(lat=lat, lon=lon, level=level)
+    # Generate cell ID and convert to token
+    return cell_id_to_token(lat_lon_to_cell_id(lat=lat, lon=lon, level=level))
 
-    # The zero token is encoded as 'X' rather than as a zero-length string. This implementation has
-    # no method of generating the 0 cell ID, so this line is mostly here for consistency with the
-    # reference implementation.
-    if cell_id == 0:  # pragma: no cover
-        return 'X'
-
-    # Convert cell ID to hex and strip any trailing zeros
-    return '{:016x}'.format(cell_id).rstrip('0')
 
+#
+# Decode functions
+#
 
 def cell_id_to_lat_lon(  # pylint: disable=too-many-locals
     cell_id: Union[int, np.uint64]
@@ -405,6 +476,7 @@ def cell_id_to_lat_lon(  # pylint: disable=too-many-locals
 
     Raises:
         TypeError: If the cell_id is not int or np.uint64.
+        ValueError: If the cell_id has invalid face bits.
 
     """
     # Check type
@@ -503,12 +575,12 @@ def cell_id_to_lat_lon(  # pylint: disable=too-many-locals
     # See s2geometry/blob/c59d0ca01ae3976db7f8abdc83fcc871a3a95186/src/s2/s2coords.h#L348-L357
     #
     # Face -> XYZ components -> indices with negation:
-    # 0    -> ( 1,  u,  v)   -> ( 2,  0,  1)
-    # 1    -> (-u,  1,  v)   -> (-0,  2,  1)
-    # 2    -> (-u, -v,  1)   -> (-0, -1,  2)
-    # 3    -> (-1, -v, -u)   -> (-2, -1, -0) <- -1 here means -1 times the value in index 1,
-    # 4    -> ( v, -1, -u)   -> ( 1, -2, -0)    not index -1
-    # 5    -> ( v,  u, -1)   -> ( 1,  0, -2)
+    # 0    -> ( 1,  u,  v)   -> ( /,  0,  1)
+    # 1    -> (-u,  1,  v)   -> (-0,  /,  1)
+    # 2    -> (-u, -v,  1)   -> (-0, -1,  /)
+    # 3    -> (-1, -v, -u)   -> (-/, -1, -0) <- -1 here means -1 times the value in index 1,
+    # 4    -> ( v, -1, -u)   -> ( 1, -/, -0)    not index -1
+    # 5    -> ( v,  u, -1)   -> ( 1,  0, -/)
     if face == 0:
         s2_point = (1, uv[0], uv[1])
     elif face == 1:
@@ -519,8 +591,10 @@ def cell_id_to_lat_lon(  # pylint: disable=too-many-locals
         s2_point = (-1, -uv[1], -uv[0])
     elif face == 4:
         s2_point = (uv[1], -1, -uv[0])
-    else:
+    elif face == 5:
         s2_point = (uv[1], uv[0], -1)
+    else:
+        raise ValueError('Cannot decode S2 cell ID with invalid face: ' + str(face))
 
     # Normalise XYZ S2Point vector
     # This section is part of the reference implementation but is not necessary when mapping
@@ -554,18 +628,76 @@ def token_to_lat_lon(token: str) -> Tuple[float, float]:
     Raises:
         TypeError: If the token is not str.
         ValueError: If the token length is over 16.
+        ValueError: If the token has invalid face bits.
+
+    """
+    # Convert to cell ID and decode to lat/lon
+    return cell_id_to_lat_lon(token_to_cell_id(token))
+
+
+#
+# Level extraction
+#
+
+def cell_id_to_level(cell_id: Union[int, np.uint64]) -> int:
+    """
+    Get the level for a S2 cell ID.
+
+    See s2geometry/blob/c59d0ca01ae3976db7f8abdc83fcc871a3a95186/src/s2/s2cell_id.h#L543-L551
+
+    Args:
+        cell_id: The S2 cell ID integer.
+
+    Returns:
+        The level of the S2 cell ID.
+
+    Raises:
+        TypeError: If the cell_id is not int or np.uint64.
+        ValueError: If the cell_id is 0.
 
     """
     # Check input
-    if not isinstance(token, str):
-        raise TypeError('Cannot decode S2 token from type: ' + str(type(token)))
+    if not isinstance(cell_id, (int, np.uint64)):
+        raise TypeError('Cannot decode S2 cell ID from type: ' + str(type(cell_id)))
+    cell_id = np.uint64(cell_id)
 
-    if len(token) > 16:
-        raise ValueError('Cannot decode S2 token with length > 16 characters')
+    if cell_id == 0:
+        # The cell ID 0 has no level
+        raise ValueError('Cannot decode invalid S2 cell ID: 0')
 
-    # Add stripped zeros
-    token = token + ('0' * (16 - len(token)))
+    # Find the position of the lowest set one bit, which will be the trailing one bit. The level is
+    # given by the max level (30) minus the floored division by two of the position of the lowest
+    # set bit.
+    #
+    # The position of the lowest set bit is found using 'count trailing zeros', which would be
+    # equivalent to the C++20 function std::countr_zero() or the ctz instruction.
+    lsb_pos = 0
+    while cell_id != 0:  # pragma: no branch
+        if cell_id & np.uint64(1):
+            break
+        lsb_pos += 1
+        cell_id >>= np.uint64(1)
 
-    # Convert to cell ID by converting hex to int and decode to lat/lon
-    cell_id = int(token, 16)
-    return cell_id_to_lat_lon(cell_id)
+    return int(_S2_MAX_LEVEL - (lsb_pos >> 1))
+
+
+def token_to_level(token: str) -> int:
+    """
+    Get the level for a S2 token.
+
+    See s2geometry/blob/c59d0ca01ae3976db7f8abdc83fcc871a3a95186/src/s2/s2cell_id.h#L543-L551
+
+    Args:
+        token: The S2 token string. Can be upper or lower case hex string.
+
+    Returns:
+        The level of the S2 token.
+
+    Raises:
+        TypeError: If the token is not str.
+        ValueError: If the token length is over 16.
+        ValueError: If the token as cell ID is 0.
+
+    """
+    # Convert to cell ID and get the level for that
+    return cell_id_to_level(token_to_cell_id(token))
